//############################################################################
//
// File Name	: 'fat.c'
// Title		: FAT16 file system driver
// Author		: Zoltan Gradwohl
// Date			: 28/07/2005
// Revised		: 28/07/2005
// Version		: 1.0
// Target MCU	: ATmega128 (2K Byte Internal SRAM required)
// Editor Tabs	: 4
//
//
// NOTE: This code is currently below version 1.0, and therefore is considered
// to be lacking in some functionality or documentation, or may not be fully
// tested.  Nonetheless, you can expect most functions to work.
//
// This code is distributed under the GNU Public License
//		which can be found at http://www.gnu.org/licenses/gpl.txt
//
//############################################################################

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <string.h>

#include "mmc.h"
#include "rprintf.h"
#include "fat.h"
#include "fatconf.h"
#include "debug.h"

// globals
volatile uint16_t  RootDirectoryRegionStart;
volatile uint32_t  DataRegionStart;
volatile uint16_t  FATRegionSize;
volatile uint16_t  RootDirectoryRegionSize;
volatile uint32_t  DataRegionSize;
volatile uint8_t   SectorsPerCluster;
volatile uint16_t  BytePerSector;
volatile uint16_t  FATRegionStart;
volatile uint16_t  NumberOfFATs;
volatile uint8_t   SectorsPerFAT;
volatile uint16_t  VolumeStart = 0;
volatile uint16_t  ReservedSectors;

uint16_t glbl_FAT_Entry = 0;	// FAT entry 
uint32_t glbl_FileSize = 0;	// filesize
uint32_t glbl_DataSector = 0;	// Sector where data starts
uint32_t glbl_Sector = 0;		// Sector which contains file size inforation
uint8_t  glbl_SubSector = 0;	// Which of the 16 32byte long SubSectors contains the file size 
uint8_t  glbl_Attributes = 0; 	// Attribute Dir, archive, hidden, system, volume ...

uint16_t FATSectorAddress_old = 0;		// Zum berprfen ob der FAT Sector schon geladen wurde	
	
#define NAMEBUFFERSIZE 32
char NameBuffer[NAMEBUFFERSIZE];
char NameBufferShort[8];
char NameBufferExt[3];		
char Name2FindBuffer[NAMEBUFFERSIZE];

union FAT16_Buffer F16;
#define F16_BUFFER_ONE (uint8_t*)(&F16.FATBuffer)
uint16_t FAT_Buf2[256];
#define F16_BUFFER_TWO (uint8_t*)(&FAT_Buf2)

#define RootEntiesCount 	 F16.BootSec.bsRootEntCnt
#define TotalNumberOfSectors F16.BootSec.bsTotSec32



//############################################################################
unsigned char fatInit(void)
//############################################################################
{
	// mmc read first sector 
	mmcRead( 0 , F16_BUFFER_ONE );
	
	//If partitions are available 
	if( F16.BootSec.bsjmpBoot == 0 )
	{
		// Number of Sectors Between the MBR and the First Sector in the Partition (32bit)
		VolumeStart = (uint16_t)F16.MBS.mbPartitionEntries[0].prStartLBA;
		// mmc read Fat16 boot sector
		mmcRead( VolumeStart , F16_BUFFER_ONE );	
	}
	
	//
	ReservedSectors 		    = F16.BootSec.bsRsvdSecCnt; 		// 6
	BytePerSector 				= F16.BootSec.bsBytesPerSec;		// 512
	SectorsPerCluster  			= F16.BootSec.bsSecPerClus;		// 4
	NumberOfFATs 		 		= F16.BootSec.bsNumFATs;			// 2
	SectorsPerFAT 		 		= F16.BootSec.bsNrSeProFAT;		// 245
	
	// Calculation Algorithms
	FATRegionStart				= VolumeStart + ReservedSectors;						// 6
	RootDirectoryRegionStart 	= FATRegionStart + (NumberOfFATs * SectorsPerFAT);		// 496	
	RootDirectoryRegionSize 	= (RootEntiesCount*32)/ BytePerSector;					// 32
	DataRegionStart 			= RootDirectoryRegionStart + RootDirectoryRegionSize;	// 528
    FATRegionSize 				= NumberOfFATs * SectorsPerFAT;							// 490
	DataRegionSize				= TotalNumberOfSectors - (ReservedSectors + FATRegionSize + RootDirectoryRegionSize);

	#ifdef DEBUG_FAT
		rprintfCRLF();
		rprintfProgStrM("OEM Name        : ");	rprintfStrLen(F16.BootSec.bsOEMName, 0, 8 );				rprintfCRLF();
		rprintfProgStrM("Bytes Per Sector: ");	rprintfNum(10,6, FALSE,' ',F16.BootSec.bsBytesPerSec);		rprintfCRLF();
		rprintfProgStrM("Sector Per Clust: ");	rprintfNum(10,6, FALSE,' ',F16.BootSec.bsSecPerClus);		rprintfCRLF();
		rprintfProgStrM("Reserved SecCnt : ");	rprintfNum(10,6, FALSE,' ',F16.BootSec.bsRsvdSecCnt);		rprintfCRLF();
		rprintfProgStrM("Number Of FAT   : ");	rprintfNum(10,6, FALSE,' ',F16.BootSec.bsNumFATs);			rprintfCRLF();
		rprintfProgStrM("Root Dir Entries: ");	rprintfNum(10,6, FALSE,' ',F16.BootSec.bsRootEntCnt);		rprintfCRLF();
		rprintfProgStrM("Total Sector 16 : ");	rprintfNum(10,6, FALSE,' ',F16.BootSec.bsTotSec16);		rprintfCRLF();
		rprintfProgStrM("Media           : ");	
			switch (F16.BootSec.bsMedia)
			{
				// TODO expand further cases
				case 0xf8:
						rprintfProgStrM("Found: Hard disk. DOS version 2.0 \r\n");
						break;
				default:
						rprintfProgStrM("Found: No Partition!\r\n");
						break;
			}
			
		rprintfProgStrM("Total Sector 32 : ");	rprintfNum(10,6, FALSE,' ',F16.BootSec.bsTotSec32);		rprintfCRLF();		
		rprintfProgStrM("Sector Per FAT  : ");	rprintfNum(10,6, FALSE,' ',F16.BootSec.bsNrSeProFAT);		rprintfCRLF();
		rprintfProgStrM("Volume Label    : ");	rprintfStrLen(F16.BootSec.bsVolLbl, 0, 11 );				rprintfCRLF();
		rprintfProgStrM("Filesystem Type : ");	rprintfStrLen(F16.BootSec.bsFileSysType, 0, 8);			rprintfCRLF();
		rprintfCRLF();
		
		rprintfProgStrM("RootDirectoryRegionStart: ");	rprintfNum(10,6, FALSE,' ',RootDirectoryRegionStart);	rprintfCRLF();
		rprintfProgStrM("DataRegionStart         : ");	rprintfNum(10,6, FALSE,' ',DataRegionStart);			rprintfCRLF();
		rprintfProgStrM("FATRegionSize           : ");	rprintfNum(10,6, FALSE,' ',FATRegionSize);				rprintfCRLF();
		rprintfProgStrM("RootDirectoryRegionSize : ");	rprintfNum(10,6, FALSE,' ',RootDirectoryRegionSize);	rprintfCRLF();
		rprintfProgStrM("DataRegionSize          : ");	rprintfNum(10,6, FALSE,' ',DataRegionSize);			rprintfCRLF();
		rprintfCRLF();
	#endif
	
	return 0;	
}



//####################################################################################
void FAT16_Search (  	char   *FilePathName,		// In: 		File name to find
						int8_t SelectMode	)	  	// In: 		Select operation: overwrite or contiue data
//####################################################################################
{
	#ifdef DEBUG_FAT
		uint32_t notused = 0;
	#endif
	int8_t ret = TRUE;

	// Reset varaibles and buffers
	FATSectorAddress_old = 0;
	glbl_FAT_Entry = 0;		// FAT entry 
	glbl_FileSize = 0;		// filesize
	glbl_DataSector = 0;	// Current data sector
	glbl_Sector = 0;		// Sector which contains file size inforation
	glbl_SubSector = 0;		// Which of the 16 32byte long SubSectors contains the file size 
	glbl_Attributes = 0; 	// Attribute Dir, archive, hidden, system, volume ...
	memset( F16_BUFFER_ONE, 0x00 , 512 );				// Clear buffer
	memset( F16_BUFFER_TWO, 0x00 , 512 );				// Clear buffer
	memset( NameBuffer, 0x00 , NAMEBUFFERSIZE );		// Clear buffer
	memset( NameBufferShort, 0x00 , 8 );				// Clear buffer
	memset( NameBufferExt, 0x00 , 3 );					// Clear buffer
	memset( Name2FindBuffer, 0x00 , NAMEBUFFERSIZE );	// Clear buffer
	
	

	#ifdef DEBUG_FAT
		rprintfProgStrM("Searching for >"); rprintfStrLen(FilePathName, 0, 64 );	
		if( SelectMode == OVERWRITE) rprintfProgStrM(" (OVERWRITE)");
		if( SelectMode == CONTINUE) rprintfProgStrM(" (CONTINUE)");
		rprintfCRLF();	
	#endif

	ret = fatSearch_File( 	FilePathName,		// In: 		File name to find
							&glbl_FAT_Entry,  	// In/Ret:  FAT entry where data starts
							&glbl_FileSize, 	// Return: 	Return filesize
							&glbl_Sector, 		// Return: 	Sector which contains file size inforation
							&glbl_SubSector,	// Return: 	Return which of the 16 32byte long SubSectors contains the file size
							&glbl_DataSector,  	// Return:  Return sector where data starts
							&glbl_Attributes, 	// Return: 	Return Attribute Dir, archive, hidden, system, volume ...
							F16_BUFFER_ONE );	// In: 		Buffer to use
	
	// empty name buffer for next use 
	memset( FilePathName, 0x00 , 64 );
	
	if( (ret == TRUE)  &&  (glbl_Attributes == ATTR_ARCHIVE) )
	{
		if( SelectMode == OVERWRITE )
		{
			// delete fat entries of current file from "glbl_FAT_Entry" on until end of file and init File
			fat_Delete( &glbl_FAT_Entry,		// In/Ret:  FAT entry where data starts 
						&glbl_FileSize,   		// Return: 	Return filesize
						&glbl_Sector,			// In: 		Sector which contains file size inforation
						&glbl_SubSector, 		// In: 		Return which of the 16 32byte long SubSectors contains the file size 
						&glbl_DataSector, 		// Return:	Sector where data starts	
						F16_BUFFER_TWO );		// In:		Select buffer to use
		}
		
		// if filesize and fat entry is zero (means that file hasn't been used until now)
		// search fat address, where data can start 
		fat_Init_File 	(	&glbl_FAT_Entry,	// In/Ret:  FAT entry where data starts
							&glbl_FileSize,    	// In: 		Filesize
							&glbl_Sector,		// In: 		Sector which contains file size inforation
							&glbl_SubSector,	// In: 		Return which of the 16 32byte long SubSectors contains the file size 
							&glbl_DataSector,	// Return:	Sector where data starts
							F16_BUFFER_ONE);	// In:		Select buffer to use
		
		#ifdef DEBUG_FAT
			rprintfCRLF();  
			rprintfProgStrM("Data starts at FAT address  : ");	rprintfNum( 10,6, FALSE,' ', glbl_FAT_Entry );	   	rprintfCRLF();
			rprintfProgStrM("Sector containing file info : ");	rprintfNum( 10,6, FALSE,' ', glbl_Sector ); 		rprintfCRLF();
			rprintfProgStrM("Data starts in sector       : ");	rprintfNum( 10,6, FALSE,' ', glbl_DataSector ); 	rprintfCRLF();
			rprintfProgStrM("FileSize [uint8_t]          : ");	rprintfNum( 10,6, FALSE,' ', glbl_FileSize );	   	rprintfCRLF();
			fatRead_File( 	&glbl_FAT_Entry, 	// In:  	Start clusters of file
							&notused, 			// In:	
							&glbl_FileSize, 	// In:	
							F16_BUFFER_ONE );	// In:		Select buffer to use
		#endif	
		
		
		if( SelectMode == CONTINUE ){
			#ifdef DEBUG_FAT 
				rprintfProgStrM("\r\n----- Searching end of file -----\r\n");
			#endif
		
			// load last sector in current file into F16_BUFFER_ONE
			fatWrite_Continue( 	&glbl_FAT_Entry, 	// In/Ret:  FAT entry where data starts
								&glbl_DataSector, 	// Return:  Current data sector address
								&glbl_FileSize,    	// Return: 	Return filesize
								F16_BUFFER_ONE );	// In:		Select buffer to use
								
			#ifdef DEBUG_FAT
				rprintfCRLF();
				rprintfProgStrM("Data ends in FAT address: ");	rprintfNum( 10,6, FALSE,' ', glbl_FAT_Entry );	   	rprintfCRLF();
				rprintfProgStrM("Data ends in sector     : ");	rprintfNum( 10,6, FALSE,' ', glbl_DataSector ); 	rprintfCRLF();
				rprintfProgStrM("FileSize                : ");	rprintfNum( 10,6, FALSE,' ', glbl_FileSize );	   	rprintfCRLF();
				rprintfProgStrM("Loading last sector of file ->"); rprintfCRLF();
				debugPrintHexTable(0x200, F16_BUFFER_ONE);
			#endif
		}
	}
	else{
		#ifdef DEBUG_FAT
			rprintfProgStrM("\r\n----- File not found -----\r\n");
		#endif
	}
	
}


//############################################################################
void fatWrite_Continue (	uint16_t *FATEntry,		// In/Ret:  FAT entry where data starts
							uint32_t *SectorAddress,	// Return:  Current data sector address
							uint32_t *FileSize,    	// Return: 	Return filesize
							uint8_t  *Buffer)			// In:		Select buffer to use
//############################################################################
{
	uint16_t FAT_Entry_Temp = 0;
	uint16_t offset = 0;
	
	// find at which sector in current cluster the data ends
	offset = *FileSize / BytePerSector; 
	offset %= SectorsPerCluster;

	// search end of file to continue to save data there
	while(*FATEntry != 0xFFFF)
	{
		FAT_Entry_Temp = *FATEntry;
		fatLoad( FATEntry, SectorAddress, F16_BUFFER_TWO ); // F16_BUFFER_TWO	
	}	
	*FATEntry = FAT_Entry_Temp;
	*SectorAddress += offset; 
	// Sector_Address_Temp contains only the address of the first sector in a cluster 
	// 'offset' contains an offset. So the exact sector address in a cluster can be determined
	// read last sector of curent file, which contains data
	mmcRead( *SectorAddress , Buffer );
}


//############################################################################
void fat_Delete 		(	uint16_t *FATEntry,		// In/Ret:  FAT entry where data starts
							uint32_t *FileSize,    	// Return: 	Return filesize
							uint32_t *Sector,			// In: 		Sector which contains file size inforation
							uint8_t  *SubSector,		// In: 		Return which of the 16 32byte long SubSectors contains the file size 
							uint32_t *DataSector,		// Return:	Sector where data starts	
							uint8_t  *Buffer)			// In:		Select buffer to use
//############################################################################
{	
	uint16_t FATSectorAddress;						
	uint16_t Temp;

	if(*FileSize != 0)
	{
		#ifdef DEBUG_FAT 
			// deleting file content by deleting FAT addresses	
			rprintfProgStrM("\r\n----- Deleting file content -----\r\n");
		#endif
		
		// Calculate which FAT-sector has to be read depending on FAT-entry  
		FATSectorAddress = 	((*FATEntry*2) / BytePerSector) + FATRegionStart;	// ((8*2)/512) + 6 = 6	
		mmcRead( FATSectorAddress , Buffer );	
		FATSectorAddress_old = FATSectorAddress;
		
		do{		
			Temp = FAT_Buf2[*FATEntry]; 	// load next FAT-address from current FAT-address
			FAT_Buf2[*FATEntry] = 0;		// clear contents in current FAT-address
			*FATEntry = Temp; 				// prepare next FAT-address
			
			// Calculate which FAT-sector has to be read depending on FAT-entry
			FATSectorAddress =  ((*FATEntry*2) / BytePerSector) + FATRegionStart;	// ((8*2)/512) + 6 = 6		
			
			// If new FAT-entry is in new FAT-sector then save old FAT-sector at first and then load new FAT-sector
			if (FATSectorAddress != FATSectorAddress_old)
			{		
				mmcWrite( FATSectorAddress_old , Buffer );	
				FATSectorAddress_old = FATSectorAddress;
				mmcRead( FATSectorAddress , Buffer );	
			}
		}
		while(*FATEntry != 0xFFFF);
		
		*FileSize = 0;
		*FATEntry = 0;
		
		// remember also to store FileSize in file information sector !!!!!
		fat_Init_File 		(	FATEntry,		// In/Ret:  FAT entry where data starts
								FileSize,    	// In: 		Filesize
								Sector,			// In: 		Sector which contains file size inforation
								SubSector,		// In: 		Return which of the 16 32byte long SubSectors contains the file size 
								DataSector,		// Return:	Sector where data starts
								Buffer);		// In:		Select buffer to use	
	}	
}


//############################################################################
void fat_Init_File 		(	uint16_t *FATEntry,		// In/Ret:  FAT entry where data starts
							uint32_t *FileSize,    	// In: 		Filesize
							uint32_t *Sector,			// In: 		Sector which contains file size inforation
							uint8_t  *SubSector,		// In: 		Return which of the 16 32byte long SubSectors contains the file size 
							uint32_t *DataSector,		// Return:	Sector where data starts
							uint8_t  *Buffer)			// In:		Select buffer to use  // not used
//############################################################################
{
	// if file is empty and if fat-entry is zero then search empty space in the FAT where data can start 
	if( (*FileSize == 0) && (*FATEntry == 0)  )
	{	
		// search next empty FAT-Entry(from zero on) where data can start
		fatWrite( FATEntry, DataSector , F16_BUFFER_TWO );
		// load file information sector in buffer
		mmcRead( *Sector , F16_BUFFER_ONE );
		// Save FAT-address where data of the current file starts into the file-name SubSector
		F16.direntries[*SubSector].deStartCluster = *FATEntry;
		// Save file size
		F16.direntries[*SubSector].deFileSize = 0;	
		//  TODO Datum 
		//F16.direntries[*SubSector].deMTime = xxx;   	// update time
		//F16.direntries[*SubSector].deMDate = yyy;		// update date
		// save file information
		mmcWrite( *Sector , F16_BUFFER_ONE );
			
		#ifdef DEBUG_FAT
			rprintfCRLF();
			rprintfProgStrM("First FAT allocated for new file : ");	rprintfNum( 10,6, FALSE,' ', *FATEntry );	   
			rprintfCRLF();	
		#endif	
	}
}


//####################################################################################
int8_t fatSearch_File (    char 	   	*FilePathName,		// In: 		File name to find
							uint16_t  	*FAT_Entry, 		// In/Ret:  FAT entry where data starts
							uint32_t  	*FileSize,    		// Return: 	Return filesize
							uint32_t  	*Sector,			// Return: 	Sector which contains file size inforation
							uint8_t   	*SubSector,   		// Return: 	Return which of the 16 32byte long SubSectors contains the file size 
							uint32_t  	*DataSector,  		// Return:  Return sector where data starts
							uint8_t   	*Attributes,  		// Return: 	Return Attribute Dir, archive, hidden, system, volume ...
							uint8_t   	*Buffer)			// In: 		Buffer to use
//####################################################################################
{
	int8_t ret = FALSE;
	uint8_t  u = 0;
	uint8_t  a = 0; 
	uint16_t i = 0;
	uint16_t k = 0;
	
	do
	{
		// fill buffer with 0x00
		memset( Name2FindBuffer, 0x00 , k );
		
		k=0;
		// search directory name in parameter "FilePathName" (i.e. My Directory/ )   0x2E = "."   0x2F = "/"
		while( ( FilePathName[i] != 0x2F ) && ( FilePathName[i] != 0x2E ) &&  ( i < NAMEBUFFERSIZE ) ) 	
		{  
			Name2FindBuffer[k++] = FilePathName[i++];
		}
		
		if(FilePathName[i] == 0x2E) // 0x2E = "."
		{
			// replace in string " "(0x20) by 0 
			for(u=0; u<NAMEBUFFERSIZE;  u++)
			{
				if(Name2FindBuffer[u] == 0x20) Name2FindBuffer[u] = 0;
			}
			memcpy( NameBufferShort,  Name2FindBuffer, 8 );
			Name2FindBuffer[k++]  	= FilePathName[i++];  		// .
			Name2FindBuffer[k] 		= FilePathName[i++];		// t    i.e .txt
			NameBufferExt[0]		= Name2FindBuffer[k++];
			Name2FindBuffer[k]  	= FilePathName[i++];		// x
			NameBufferExt[1]		= Name2FindBuffer[k++];
			Name2FindBuffer[k]  	= FilePathName[i];			// t
			NameBufferExt[2]		= Name2FindBuffer[k];	
		}
		
		if(i < NAMEBUFFERSIZE)
		{	
			// search 100 fat tables
			do
			{
				ret = fatGetDirEntry( 	Name2FindBuffer,// In: 		String with name to find in sector
										FAT_Entry,  	// In/Ret:  FAT entry where data starts
										FileSize, 		// Return: 	Return filesize
										Sector, 		// Return: 	Sector which contains file size inforation
										SubSector, 		// Return: 	Return which of the 16 32byte long SubSectors contains the file size 
										DataSector,  	// Return:  Return sector where data starts
										Attributes, 	// Return: 	Return Attribute Dir, archive, hidden, system, volume ...
										Buffer );		// In: 		Buffer to use	
			}	
			while( (a++<100)  &&  (ret != TRUE) );
			
			if(ret == FALSE) return( FALSE );
			
			// skip "/" in FilePathName
			i++;
		}
	}
	// search whole buffer for dir names
	while( i < NAMEBUFFERSIZE );	
	
	
	return( ret ); // No file or dir found
}


//############################################################################
// Search a cluster for file name
// If "root dir entry" is selected (*FAT_Entry = READROOTDIR = 0) search all root dir entry secors. The length 
// of this area is stored in RootDirectoryRegionSize
int8_t fatGetDirEntry (	char     *Name2Find,	// In: 		String with name to find in sector
							uint16_t *FAT_Entry,	// In/Ret:  FAT entry where data starts
							uint32_t *FileSize,    // Return: 	Return filesize
							uint32_t *Sector,		// Return: 	Sector which contains file size inforation
							uint8_t  *SubSector,   // Return: 	Return which of the 16 32byte long SubSectors contains the file size 
							uint32_t *DataSector,  // Return:  Return sector where data starts
							uint8_t  *Attributes,  // Return: 	Return Attribute Dir, archive, hidden, system, volume ...
							uint8_t  *Buffer)		// In: 		Buffer to use
//############################################################################
{
	int16_t i = 0;
	int16_t NoSecorToSerach;
	uint8_t nr_LFN_entries;
	uint8_t order = 0;
	uint8_t k = 0;
	uint8_t p = 0;
	uint8_t m = 0;
	uint8_t q = 0;
	uint8_t u = 0;
	

	// if READROOTDIR then search root directory region ( 496... ) // READROOTDIR = 0
	// *FAT_Entry == 0 is treated as root directory region start address
	if(*FAT_Entry == READROOTDIR){ 
		// RootDirectoryRegionStart --> RootDirectoryRegionSize [sectors]
		*Sector = RootDirectoryRegionStart;  
		NoSecorToSerach = RootDirectoryRegionSize; 	// search RootDirectoryRegionSize entries (32)
	//	#ifdef DEBUG_FAT
	//		rprintfProgStrM("----- Searching in root dir entry -----");	rprintfCRLF();			   
	//	#endif
	}
	else{ 
		fatLoad( FAT_Entry, Sector, F16_BUFFER_TWO );
		NoSecorToSerach = SectorsPerCluster;     	// search SectorsPerCluster entries (4)
	}
	
	do
	{
		// search all sectors in selected cluster
		// Read data from cluster
		mmcRead( *Sector + q, Buffer );

		i=0;m=0;p=0;
		do 	
		{	
			// fill buffer with 0x00 
			memset( NameBuffer, 0x00 , NAMEBUFFERSIZE );
			
			// If not a deleted entry (0xE5)
			if( F16.direntries[i].deName[0] != SLOT_DELETED )
			{	
				switch( F16.direntries[i].deAttributes )
				{
					// If long filename entry found (0x0F) ---------------------------------------------------------------------------- 
					case ATTR_LONG_FILENAME: 							// deleted if 7th bit is set in .weCnt?
					{
						// We have a long name entry
						// Max (16) 20(10100b) "subSectors" in a Sector(512 byte). 
						// Each Subsector contains max 13 UTF-16 characters in 26 bytes.	
						nr_LFN_entries = (uint8_t)( F16.longdirentries[i].weCnt & 0x1F ) ;	
						
						for( m=0; m < nr_LFN_entries; m++ ){
							if((i+m) == 16){				// if LFN goes over the length of a sector, load next sector
								// if searching in the root dir entry, no information in fat table is available
								// the root dir entry lays in a specified area (see Boot Sector ~> BIOS Parameter Block) 
								if( (q == SectorsPerCluster-1 ) && (*FAT_Entry != READROOTDIR) ){ 
									// if q=3 (4 sectors per cluster) then load next fat and...
									fatLoad( FAT_Entry , Sector, F16_BUFFER_TWO );
									//... reset sector counter to zero
									q=0;
								}
								// load new secor
								mmcRead( ++q + *Sector , Buffer);
								i-=16;						// start to read the next sector from zero position: (i+m)=0
							}
							p=0;
							order = (nr_LFN_entries - m - 1) * 13;	// load SubSector-data(each one 32Byte) in correct order in NameBuffer;
							for(k=0;k<5;k++)	NameBuffer[order + p++] = F16.longdirentries[i+m].wePart1[k*2]; 	// copy first part
							for(k=0;k<6;k++)	NameBuffer[order + p++] = F16.longdirentries[i+m].wePart2[k*2]; 	// second part 
							for(k=0;k<2;k++)	NameBuffer[order + p++] = F16.longdirentries[i+m].wePart3[k*2]; 	// and third part 	
						} 
						if((i+m) == 16){	// if LFN goes over the length of a sector, load next sector 
							// if searching in the root dir entry, no information in fat table is available
							// the root dir entry lays in a specified area (see Boot Sector ~> BIOS Parameter Block) 
							if( (q == SectorsPerCluster-1) && (*FAT_Entry != READROOTDIR) ){ 
								// if q=3 (4 sectors per cluster) then load next fat and...
								fatLoad( FAT_Entry , Sector, F16_BUFFER_TWO );
								//... reset sector counter to zero
								q=0;
							}
							// load new secor
							mmcRead( ++q + *Sector , Buffer);		
							i-=16;				// start to read the next sector from zero position: (i+m)=0
						}
						
						// replace empty data(0xFF) in file name by zeros 
						// this is a necessary requirement to be able to use the string compare function
						for(u=0; u< NAMEBUFFERSIZE;  u++) 
						{
							if(NameBuffer[u] == 0xFF) NameBuffer[u] = 0;
						}
						
						if(!strncmp( NameBuffer, Name2Find , 32 ) ){
							*FAT_Entry 	= F16.direntries[i+m].deStartCluster;
							*FileSize  	= F16.direntries[i+m].deFileSize;
							*Attributes = F16.direntries[i+m].deAttributes;
							*DataSector	= DataRegionStart + ( ( *FAT_Entry - 2 ) * SectorsPerCluster ) ;
							*SubSector 	= i+m;
							*Sector    += q;
							
							#ifdef DEBUG_FAT
								PrintDOSLine(	NameBuffer, 
												F16.direntries[i+m].deCDate,
												F16.direntries[i+m].deCTime,
												F16.direntries[i+m].deFileSize,
												F16.direntries[i+m].deAttributes);
							#endif
							
							return (TRUE);
						}
							
						i=i+m;	// update current position in cluster
						break;
					}
					// if short filename entry found ----------------------------------------------------------------------------------
					case ATTR_DIRECTORY:
					{
						if( F16.direntries[i].deName[0] != SLOT_DNt_Entry )  	// if dir name beginns with '.'
						{	
							// copy data in NameBuffer
							for(k=0;k<8;k++)   NameBuffer[k] = F16.direntries[i].deName[k];
							
							// replace "space"(0x20) in file name by zeros 
							// this is a necessary requirement to be able to use the string compare function
							for( u=0; u<8;  u++ )
							{
								if( NameBuffer[u] == 0x20 ) NameBuffer[u] = 0;
							}
							
							// if name match with input dir get dir information from buffer
							if( !strncmp( NameBuffer, Name2Find , 32 ) ){
								
								*FAT_Entry 	= F16.direntries[i].deStartCluster;
								*FileSize  	= F16.direntries[i].deFileSize;
								*Attributes = F16.direntries[i].deAttributes;
								*DataSector	= DataRegionStart + ( ( *FAT_Entry - 2 ) * SectorsPerCluster ) ; ;
								*SubSector 	= i;
							    *Sector    += q;
								
								#ifdef DEBUG_FAT
									PrintDOSLine(	NameBuffer, 
													F16.direntries[i].deCDate,
													F16.direntries[i].deCTime,
													F16.direntries[i].deFileSize,
													F16.direntries[i].deAttributes);
								#endif
								
								return (TRUE);
							}
						}
						break; 
					}	
					// if data entry found --------------------------------------------------------------------------------------------
					case ATTR_ARCHIVE:
					{
						// copy data in NameBuffer
						for(k=0;k<8;k++)   NameBuffer[k] = F16.direntries[i].deName[k];
						NameBuffer[8] = '.';
						for(k=0;k<3;k++)  NameBuffer[k+9]  = F16.direntries[i].deExtension[k];
						
						// replace zeros in file name "space"(0x20) 
						// this is a necessary requirement to be able to use the string compare function
						for(u=0; u<8;  u++)
						{
							if( NameBufferShort[u] == 0x00 ) NameBufferShort[u] = 0x20;
						}
						for(u=0; u<3;  u++)	
						{
							if( NameBufferExt[u] == 0x00 )   NameBufferExt[u] = 0x20;	
						}	
						
						strupr(NameBufferShort); // convert string to upper case
						strupr(NameBufferExt);	 // convert string to upper case
						
						// if name and extension match with input file get file information from buffer
						if	( (!strncmp( NameBufferShort, F16.direntries[i].deName,    8 ) )   &&
							  (!strncmp( NameBufferExt, F16.direntries[i].deExtension, 3 ) ) 
							)
						{
							*FAT_Entry 	= F16.direntries[i].deStartCluster;
							*FileSize  	= F16.direntries[i].deFileSize;
							*Attributes = F16.direntries[i].deAttributes;
							*DataSector	= DataRegionStart + ( ( *FAT_Entry - 2 ) * SectorsPerCluster ) ;
							*SubSector 	= i;
							*Sector    += q;
							
							#ifdef DEBUG_FAT
								PrintDOSLine(	NameBuffer, 
												F16.direntries[i].deCDate,
												F16.direntries[i].deCTime,
												F16.direntries[i].deFileSize,
												F16.direntries[i].deAttributes);
							#endif
							
							return (TRUE);
						}
						break;	
					}
					default:	
						break;
				}	
			}
			i++;	
		}while( (F16.direntries[i].deAttributes != ATTR_NORMAL)  &&  (i != 16) );	// 0 in de->deName[0] if no more entries	
		
		q++;	
	}while( q < NoSecorToSerach );
	
	
	// save actual sector address
	*Sector += q;
	// if no file found start next xearch from fat entry 1 (*FAT_Entry == 0 would again read all root dir entries)
	*FAT_Entry = 1;
	// file or dir not found
	return (FALSE);
} 


//############################################################################
void fatLoad (	uint16_t *FATEntry, 		// In:     	Angabe Start cluster , Rckgabe des nchsten FAT- Eintrags auf dem das aktuelle FAT-Eintrag zeigt
				uint32_t *Adr_FirstSectorOfCluster,// Return:	Rckgabe Adresse des ersten Sectors des gewhlten neuen Clusters
				uint8_t  *Buffer) 			  		// In: 		Buffer to use
//############################################################################
{	
	uint16_t FATSectorAddress;						// FAT Sector Adresse			

	// calculate sector that has to be read
	FATSectorAddress = 	((*FATEntry*2) / BytePerSector) + FATRegionStart;	// ((8*2)/512) + 6 = 6		

	// if sought sector is a new sector then load this sector 
	if (FATSectorAddress != FATSectorAddress_old){
		FATSectorAddress_old = FATSectorAddress;
		mmcRead( FATSectorAddress , Buffer );		
	}

	// first sector address of actual cluster
	*Adr_FirstSectorOfCluster = DataRegionStart + ( ( *FATEntry - 2 ) * SectorsPerCluster ) ;
	// next FAT number of of this file
	*FATEntry = FAT_Buf2[*FATEntry]; 
}


//############################################################################
void  fatRead_File (	uint16_t *FATEntry, 		// In: 
						uint32_t *Sector_Address,	// In:
						uint32_t *FileSize, 		// In:					
						uint8_t  *Buffer)			// In: 		Buffer to use 		
//############################################################################
{
	*Sector_Address = 0;
	uint8_t c;
	uint16_t NextCluster_save = *FATEntry;
	
	if( *FileSize != 0)
	{
		#ifdef DEBUG_FAT 
			rprintfProgStrM("\r\n----- Read file content-----\r\n");
		#endif
		
		while(*FATEntry != 0xFFFF)
		{
			#ifdef DEBUG_FAT
				rprintfProgStrM("Current cluster: "); rprintfNum(10,6, FALSE,' ', *FATEntry );
				rprintfProgStrM(" Current sector: ");rprintfNum(10,10, FALSE,' ', *Sector_Address ); rprintfCRLF();
			#endif
			
			fatLoad( FATEntry , Sector_Address, F16_BUFFER_TWO ); // F16_BUFFER_TWO
			
			#ifdef DEBUG_FAT
				rprintfProgStrM("Next cluster   : ");rprintfNum(10,6, FALSE,' ', *FATEntry );
				rprintfProgStrM(" Next sector   : ");rprintfNum(10,10, FALSE,' ', *Sector_Address );rprintfCRLF();
			#endif
			
			c=0;
			while(c < SectorsPerCluster)
			{
				mmcRead( *Sector_Address + c , Buffer);
				
				#ifdef DEBUG_FAT
					// print buffer contents
					rprintfCRLF();
					rprintfProgStrM("Sector   : ");	rprintfNum( 10,6, FALSE,' ', *Sector_Address+c );	   rprintfCRLF();
					rprintfProgStrM("Next FAT : ");	rprintfNum( 10,6, FALSE,' ', *FATEntry );	   rprintfCRLF();
					debugPrintHexTable(0x200, Buffer);
				#endif
				
				c++;
			}	
		}
	}
	#ifdef DEBUG_FAT
	else{
		rprintfProgStrM("\r\n----- File is empty -----\r\n");
	}
	#endif
	
	*FATEntry = NextCluster_save; // restore cluster number
}


//############################################################################
void FAT16_Write_Data (  	uint32_t Data )	  	// In: 	Data to be stored on MMC		
//############################################################################
{
	fatWrite_Data ( &Data, 
					&glbl_FAT_Entry, 
					&glbl_DataSector, 
					&glbl_FileSize,  
					F16_BUFFER_ONE );	
}


//############################################################################
void fatWrite_Data (	uint32_t *data,			// In: 
						uint16_t *FATEntry,		// In/Ret: 
						uint32_t *DataSector,		// In/Ret: 
						uint32_t *FileSize, 		// In/Ret:
						uint8_t  *Buffer)			// In: 		Buffer to use
//############################################################################
{
	F16.FATData[ (*FileSize/4) % 128 ] = *data;  	// write data in buffer
	(*FileSize) += 4;								// file size counter (+4 byte because 32 bit data )
	
	if( ((*FileSize/4) % 128) == 0 )				// SectorsPerCluster = 4
	{
		#ifdef DEBUG_FAT
			rprintfProgStrM("--FileSize       : ");		rprintfNum( 10,6, FALSE,' ', *FileSize );	 rprintfCRLF();
			rprintfProgStrM("--DataSector     : ");		rprintfNum( 10,6, FALSE,' ', *DataSector ); rprintfCRLF();
		#endif
		
		mmcWrite( (*DataSector)++ , Buffer );
		
		if( *DataSector % SectorsPerCluster == 0 ) 
		{	
			#ifdef DEBUG_FAT      
				rprintfProgStrM("Actual Cluster   : "); rprintfNum(10,10, FALSE,' ', *FATEntry ); 	rprintfCRLF();
				rprintfProgStrM("Data Sector      : "); rprintfNum(10,10, FALSE,' ', *DataSector ); 	rprintfCRLF();	
			#endif
			fatWrite( FATEntry, DataSector, F16_BUFFER_TWO );
			#ifdef DEBUG_FAT
				rprintfProgStrM("Next Cluster     : "); rprintfNum(10,10, FALSE,' ', *FATEntry ); 	rprintfCRLF();
				rprintfProgStrM("Sector Address   : "); rprintfNum(10,10, FALSE,' ', *DataSector ); 	rprintfCRLF();
			#endif
		}
	}
}


//############################################################################
void FAT16_Finish_Data( void )					//		
//############################################################################
{	
	// save last sector of current file
	mmcWrite( glbl_DataSector, F16_BUFFER_ONE );
	
	// save file size in file information sector
	mmcRead( glbl_Sector , F16_BUFFER_ONE );
	F16.direntries[glbl_SubSector].deFileSize = glbl_FileSize;
	//  TODO Datum update
	//F16.direntries[glbl_SubSector].deMTime = xxx;   	// last update time
	//F16.direntries[glbl_SubSector].deMDate = yyy;		// last update date
	mmcWrite( glbl_Sector , F16_BUFFER_ONE );
	
	// reload last sector of current file
	mmcRead( glbl_DataSector , F16_BUFFER_ONE );
	
	#ifdef DEBUG_FAT      
		rprintfProgStrM("glbl_DataSector : ");	rprintfNum( 10,6, FALSE,' ', glbl_DataSector );	rprintfCRLF();
		rprintfProgStrM("glbl_Sector     : ");	rprintfNum( 10,6, FALSE,' ', glbl_Sector );	   	rprintfCRLF();
		rprintfProgStrM("glbl_FileSize   : ");	rprintfNum( 10,6, FALSE,' ', glbl_FileSize );	   	rprintfCRLF();
		rprintfProgStrM("\r\n----- File finished -----\r\n\r\n"); 
	#endif
}


//############################################################################
void fatWrite(	uint16_t *FATEntry, 				// In/Ret: 	Angabe Start cluster , Rckgabe des nchsten FAT- Eintrags auf dem das aktuelle FAT-Eintrag zeigt
				uint32_t *Adr_FirstSectorOfCluster,// Return: 	Rckgabe Addresse des ersten Sectors des gewhlten Clusters
				uint8_t  *Buffer)					// In: 		Buffer to use
//############################################################################
{	
	uint16_t FATSectorAddress;						// FAT Sector Addresse			
	uint16_t FATEntry_start = *FATEntry;
	uint16_t FATSectorAddress_start;


	// Berechnung des Sectors der gelesen werden mu
	FATSectorAddress = 	((*FATEntry*2) / BytePerSector) + FATRegionStart;	// ((8*2)/512) + 6 = 6	 //((800*2)/512) + 6 = 9
	FATSectorAddress_start = FATSectorAddress;
		
	// Wenn gesuchter FAT-Eintag im neuer FAT-Sector Bereich dann neuen Sector laden
	if (FATSectorAddress != FATSectorAddress_old)
	{
		FATSectorAddress_old = FATSectorAddress;
		mmcRead( FATSectorAddress , Buffer );		
	}

	
	// Find next empty cluster
	while( FAT_Buf2[(*FATEntry) % 256 ] != 0x0000 )
	{
		(*FATEntry)++;
		if( *FATEntry % 256 == 0 ){
			mmcRead( FATSectorAddress++ , Buffer );
			// TODO return "no space" if end of fat table reached
		}	
	}
	// If empty FAT-Entry found, mark this field as end of document(0xFFFF)
	FAT_Buf2[(*FATEntry) % 256] = 0xFFFF;
	
	
	// If file is empty, do not overwrite the previous fat address
	// change instead the fat address in the Directory Entry Structure (32byte long) -> Starting cluster
	if(glbl_FileSize != 0x0000){
		// If no new Fat-Sector has been loaded during search for empty cluster (0x0000)...
		if (FATSectorAddress == FATSectorAddress_old)
		{
			#ifdef DEBUG_FAT     
				rprintfProgStrM("FAT Table Entry  : "); rprintfNum(10,10, FALSE,' ', *FATEntry ); rprintfCRLF();
			#endif
			// ...overwrite the old cluster address with the address of the new cluster
			FAT_Buf2[FATEntry_start % 256] = *FATEntry;
		}
		// Save the entry on data medium
		mmcWrite( FATSectorAddress , Buffer );
		
		// 
		if(FATSectorAddress != FATSectorAddress_old)
		{
			FATSectorAddress_old = FATSectorAddress;
			// If we use a cluster that isn't the cluster in which we started to search for an empty place
			// then read the startig sector, to actualize the postion of the 0xFFFF in the old FAT-entry.
			mmcRead( FATSectorAddress_start , Buffer );
			// Overwrite the old cluster address with the address of the new cluster
			FAT_Buf2[FATEntry_start % 256] = *FATEntry;
			// Save the entry on data medium
			mmcWrite( FATSectorAddress_start , Buffer );
		}
	}
	else{
		// Save the entry on data medium
		mmcWrite( FATSectorAddress , Buffer );
	}
	
	// Return first sector-address of new cluster
	*Adr_FirstSectorOfCluster = DataRegionStart + ( ( *FATEntry - 2 ) * SectorsPerCluster ) ;	
	
}


#ifdef DEBUG_FAT
//############################################################################
// Creates DOS like output: 08.07.2005    15:53   <DIR>    My New Directory
// Creates DOS like output: 08.07.2005    15:54        541 Textdokument.txt
void PrintDOSLine ( char 		*DirNameAddress, 	// In: 		
					uint16_t 	CreateDate, 		// In: 		
					uint16_t 	CreateTime,			// In: 		
					uint32_t 	File_Size,			// In: 		
					uint8_t 	Dir_Or_Data)		// In: 		Buffer to use
//############################################################################
{
	rprintfNum(10,2, FALSE,'0',( CreateDate & DD_DAY_MASK ) >> DD_DAY_SHIFT ); 			// Date
	rprintfProgStrM(".");
	rprintfNum(10,2, FALSE,'0',( CreateDate & DD_MONTH_MASK ) >> DD_MONTH_SHIFT ); 		// Month
	rprintfNum(10,5, FALSE,'.',( (CreateDate & DD_YEAR_MASK) >> DD_YEAR_SHIFT ) + 1980);	// Year
	rprintfProgStrM("    ");
	rprintfNum(10,2, FALSE,'0',( CreateTime & DT_HOURS_MASK ) >> DT_HOURS_SHIFT ); 		// Hours
	rprintfProgStrM(":");
	rprintfNum(10,2, FALSE,'0',( CreateTime & DT_MINUTES_MASK ) >> DT_MINUTES_SHIFT );		// Min
	if(Dir_Or_Data == ATTR_DIRECTORY) {
		rprintfProgStrM("   <DIR>    ");
	}
	if(Dir_Or_Data == ATTR_ARCHIVE){ 
		rprintfProgStrM(" ");
		rprintfNum(10,10, FALSE,' ', File_Size ); 
		rprintfProgStrM(" ");
	}
	rprintfStrLen(DirNameAddress, 0, 32 );					// Dir or file name
	rprintfCRLF();		// New line
}
#endif
