/*! \file spi.c \brief SPI interface driver. */
//############################################################################
//
// File Name	: 'spi.c'
// Title		: SPI interface driver
// Author		: Pascal Stang - Copyright (C) 2000-2002
// Modified     : Zoltan Gradwohl 06/29/2005
// Created		: 11/22/2000
// Revised		: 06/06/2002
// Version		: 0.6
// Target MCU	: Atmel AVR series
// Editor Tabs	: 4
//
// NOTE: This code is currently below version 1.0, and therefore is considered
// to be lacking in some functionality or documentation, or may not be fully
// tested.  Nonetheless, you can expect most functions to work.
//
// This code is distributed under the GNU Public License
//		which can be found at http://www.gnu.org/licenses/gpl.txt
//
//############################################################################

#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>

#include "spi.h"

// Define the SPI_USEINT key if you want SPI bus operation to be
// interrupt-driven.  The primary reason for not using SPI in
// interrupt-driven mode is if the SPI send/transfer commands
// will be used from within some other interrupt service routine
// or if interrupts might be globally turned off due to of other
// aspects of your program
//
// Comment-out or uncomment this line as necessary
//#define SPI_USEINT

// global variables
volatile uint8_t spiTransferComplete;

// SPI interrupt service handler
#ifdef SPI_USEINT
SIGNAL(SIG_SPI)
{
	spiTransferComplete = TRUE;
}
#endif

//############################################################################
// access routines
void spiInit()
//############################################################################
{
#ifdef __AVR_ATmega128__
	// setup SPI I/O pins
	PORTB |= (1<<PB1);	// set SCK hi
	DDRB |= (1<<PB1);	// set SCK as output
	DDRB &= ~(1<<PB3);	// set MISO as input
	DDRB |= (1<<PB2);	// set MOSI as output
	DDRB |= (1<<PB0);	// SS must be output for Master mode to work
#else
	// setup SPI I/O pins (ATMEGA8)
	PORTB |= (1<<PB5);	// set SCK hi
	DDRB |= (1<<PB5);	// set SCK as output
	DDRB &= ~(1<<PB4);	// set MISO as input
	DDRB |= (1<<PB3);	// set MOSI as output
	DDRB |= (1<<PB2);	// SS must be output for Master mode to work
#endif
	
	// setup SPI interface :
	// master mode
	SPCR |= (1<<MSTR);
	// clock = f/4
//	SPCR &= ~(1<<SPR0);
//	SPCR &= ~(1<<SPR1);
	// clock = f/16
	SPCR &= ~(1<<SPR0);
	SPCR |= (1<<SPR1);
	// select clock phase positive-going in middle of data
	SPCR &= ~(1<<CPOL);
	// Data order MSB first
	SPCR &= ~(1<<DORD);
	// enable SPI
	SPCR |= (1<<SPE);
		
	// some other possible configs
	//SPCR = (1<<MSTR)|(1<<SPE)|(1<<SPR0);
	//SPCR = (1<<CPHA)|(1<<CPOL)|(1<<MSTR)|(1<<SPE)|(1<<SPR0)|(1<<SPR1);
	//SPCR = (1<<CPHA)|(1<<MSTR)|(1<<SPE)|(1<<SPR0);
	
	// clear status
	SPSR;
	spiTransferComplete = TRUE;

	// enable SPI interrupt
	#ifdef SPI_USEINT
	sbi(SPCR, SPIE);
	#endif
}
/*
void spiSetBitrate(uint8_t spr)
{
	outb(SPCR, (inb(SPCR) & ((1<<SPR0)|(1<<SPR1))) | (spr&((1<<SPR0)|(1<<SPR1)))));
}
*/

//############################################################################
// Send one byte 
void spiSendByte(uint8_t data)
//############################################################################
{
	// send a byte over SPI and ignore reply
	#ifdef SPI_USEINT
		while(!spiTransferComplete); 	
	#else
	//	while(SPSR & (1<<SPIF));		// '!' Fehler in der Library????
	#endif

	spiTransferComplete = FALSE;
	SPDR = data;

	// Wait for transmission complete, util SPIF is set
	while(!(SPSR & (1<<SPIF)));
	
}

//############################################################################
// Shift one byte out SPDR and read data shifted in meanwhile
uint8_t spiTransferByte(uint8_t data)
//############################################################################
{
	// make sure interface is idle
	#ifdef SPI_USEINT
		while(!spiTransferComplete); 	
	#else
	//	while(SPSR & (1<<SPIF));		// '!' Fehler in der Library????
	#endif

	// send the given data
	spiTransferComplete = FALSE;
	SPDR = data;
	
	// wait for transfer to complete
	#ifdef SPI_USEINT
		while(!spiTransferComplete);
	#else
		while(!(SPSR & (1<<SPIF)));
		// *** reading of the SPSR and SPDR are crucial
		// *** to the clearing of the SPIF flag
		// *** in non-interrupt mode
		//SPDR;
		// set flag
		spiTransferComplete = TRUE;
	#endif
	// return the received data
	return SPDR;
}

//############################################################################
// Shift two bytes out SPDR and read data shifted in meanwhile
uint16_t spiTransferWord(uint16_t data)
//############################################################################
{
	uint16_t rxData = 0;

	// send MS byte of given data
	rxData = (spiTransferByte((data>>8) & 0x00FF))<<8;
	// send LS byte of given data
	rxData |= (spiTransferByte(data & 0x00FF));

	// return the received data
	return rxData;
}
